unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, WndSub32, Menus;

type
  TMyShape = (msEllipse, msRect, msTriangle, msWeird, msNone);
  TForm1 = class(TForm)
    Subc1: TSubclass32;
    Start: TButton;
    lb1: TListBox;
    Button2: TButton;
    Label1: TLabel;
    pm: TPopupMenu;
    ME: TMenuItem;
    MR: TMenuItem;
    MT: TMenuItem;
    Stop: TButton;
    N1: TMenuItem;
    MN: TMenuItem;
    N2: TMenuItem;
    RS: TMenuItem;
    procedure StartClick(Sender: TObject);
    procedure DrawWindowButton(down: Boolean);
    procedure Button2Click(Sender: TObject);
    procedure SetButtonRect;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure MakeWindowHole(shape: TMyShape);
    procedure StopClick(Sender: TObject);
    procedure CheckForExplorer;
    procedure LaunchExplorer;
    procedure FormCreate(Sender: TObject);
    procedure Cleanup;
    {$IFDEF VER100}    //if D3
    procedure Subc1MsgReceived(const Wnd: Integer; var Msg: Cardinal;
      var wParam, lParam: Integer; const ProcessId: Integer; rmFlag: TWhen;
      var return: Integer; var CallDefWndproc: Boolean);
    {$ENDIF}
    {$IFDEF VER90}   //if D2                                            //for all Win32 Delphi version compatiblity
    procedure Subc1MsgReceived(const Wnd: Integer; var Msg: Cardinal;
      var wParam, lParam: Longint; const ProcessId: Integer; rmFlag: TWhen;
      var return: Longint; var CallDefWndproc: Boolean);
    {$ENDIF}
    {$IFDEF VER120}   //if D4
    procedure Subc1MsgReceived(const Wnd: HWND; var Msg: Cardinal;
       var wParam, lParam: Integer; const ProcessId: Cardinal; rmFlag: TWhen;
       var return: Integer; var CallDefWndproc: Boolean);
    {$ENDIF}
  private
    wrect,
    btnRect: TRect;
    ySize,
    xFrame,
    yFrame,
    btnLeft,
    btnRight: Integer;
    HolePunched: Boolean;
    rgn, rgn2: HRGN;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

const
  HTMINE = 50;     //constant for the window procedure to return when mouse is over the 'Slice it up!' button
  btnWidth = 100;  //'Slice it up!' button width
  btnCaption = 'Slice it up!';  //button caption

{$R *.DFM}

//creates the hole in the window
procedure TForm1.MakeWindowHole(shape: TMyShape);
var arr: array[0..9] of TPoint;
    i: Integer;
begin
  if shape = msNone then begin
    SetWindowRgn(Subc1.WindowHandle, 0, True);
    DeleteObject(rgn);
  end
  else begin      //set up all the region stuff
    SetButtonRect;
    rgn := CreateRectRgn(0, 0, screen.width + xFrame, screen.height + yFrame);
    with wrect do
      case shape of
        msEllipse:  rgn2 := CreateEllipticRgn(50, 50, right-left-50, bottom-top-50);
        msRect:     rgn2 := CreateRectRgn(50, 50, right-left-50, bottom-top-50);
        msTriangle: begin
                      arr[0] := Point((right-left) div 2, 50);       //top
                      arr[1] := Point(50, bottom-top-50);            //bottom left
                      arr[2] := Point(right-left-50, bottom-top-50); //bottom right
                      rgn2 := CreatePolygonRgn(arr, 3, WINDING);
                    end;
        msWeird: begin
                   for i := 0 to 9 do
                     arr[i] := Point(random(right-left), random(bottom-top));  //some random points
                   rgn2 := CreatePolygonRgn(arr, 10, WINDING);
                 end;
      end;
    CombineRgn(rgn, rgn, rgn2, RGN_XOR);   //combine the regions
    DeleteObject(rgn2);
    if not SetWindowRgn(Subc1.WindowHandle, rgn, True) then  //stick the hole
      MessageBeep(0);
  end;
end;

//returns the Explorer handle
function ExplorerHandle: hWnd;
begin
  result := FindWindow('ExploreWClass',nil);
end;

//Launches Explorer with proper call to CreateProcess
procedure TForm1.LaunchExplorer;
var pi: TProcessInformation;
    si: TStartupInfo;
begin
  FillMemory(@si, sizeof(si), 0);
  si.cb := sizeof(si);
  CreateProcess(nil, 'explorer', nil, nil, False, NORMAL_PRIORITY_CLASS, nil, nil, si, pi);
  WaitForInputIdle(pi.hProcess, INFINITE);
  CloseHandle(pi.hProcess);
  CloseHandle(pi.hThread);
end;

procedure TForm1.CheckForExplorer;
begin
  if (ExplorerHandle = 0) and
     (MessageBox(handle, 'This demo needs Windows Explorer running, launch it now?',
                             'Confirm', MB_YESNO) = IDYes) then
    LaunchExplorer;
end;

//sets and updates the button rect (in case window position changes, etc)
procedure TForm1.SetButtonRect;
begin
  ySize := GetSystemMetrics(SM_CYSIZE);        //get height of the caption buttons
  xFrame := GetSystemMetrics(SM_CXFIXEDFRAME); //frame sizes
  yFrame := GetSystemMetrics(SM_CYFIXEDFRAME);
  GetWindowRect(Subc1.windowhandle, wrect);
  with wrect do begin
    btnLeft := ((right-left) div 2)-(btnWidth div 2);
    btnRight := btnLeft+btnWidth;              //setting the button dimensions
    btnRect := Rect(btnLeft, yFrame+3, btnRight, ySize+2);
  end;
end;

//Draws the 'Slice it up!' button
procedure TForm1.DrawWindowButton(down: Boolean);
begin
  SetButtonRect;
  Canvas.handle := GetWindowDC(Subc1.windowhandle);
  with Canvas do begin
    Font := label1.Font;
    if down then
      DrawEdge(handle, btnRect, EDGE_SUNKEN, BF_RECT)  //if button is pressed down, draw sunken
    else
      DrawEdge(handle, btnRect, EDGE_RAISED, BF_RECT); //if button is up, draw raised
    Brush.Color := clBtnFace;
    with btnRect do FillRect(Rect(left+2, top+2, right-2, bottom-2));  //fill with gray color
    if down then begin
      inc(btnRect.left,2);
      inc(btnRect.top,2);
    end;
    brush.style := bsClear;
    DrawText(handle, btnCaption, length(btnCaption), btnRect, DT_CENTER or DT_SINGLELINE or DT_VCENTER); //draw caption
    ReleaseDC(Subc1.windowhandle, handle);
  end;
end;

procedure TForm1.StartClick(Sender: TObject);
begin
  CheckForExplorer;
  Messagebox(handle, 'About to start subclass, click OK here and play around with the changes to Windows Explorer', 'Info', MB_ICONINFORMATION);
  Subc1.WindowHandle := ExplorerHandle;  //assign explorer handle
  Subc1.start := True;                   //start subclass
  if not Subc1.start then exit;          //if not successful, then exit
  DrawWindowButton(False);               //draw button up
  InsertMenu(GetMenu(Subc1.WindowHandle), 4, MF_BYPOSITION or MF_STRING or MF_POPUP, pm.handle, '&SAMPLE MENU (TSubclass32)');
     //insert the popup menu using the caption specified in the last param of this function
  DrawMenuBar(Subc1.WindowHandle);       //redraw menu bar
  HolePunched := False;
  ShowWindow(Subc1.WindowHandle, SW_SHOWNORMAL);
  SetForegroundWindow(Subc1.WindowHandle);
  Start.enabled := False;
  Stop.enabled := True;
end;

//called from a couple places, just removes the hole and removes the menu
procedure TForm1.Cleanup;
var str: array[0..255] of char;
    menu: hMenu;
begin
  MakeWindowHole(msNone);  //remove hole

  menu := GetMenu(Subc1.WindowHandle);
  GetMenuString(menu, 4, str, 255, MF_BYPOSITION);
  if StrPas(str) = '&SAMPLE MENU (TSubclass32)' then
    RemoveMenu(menu, 4, MF_BYPOSITION)     //remove menu if Win95
  else begin
    GetMenuString(menu, 0, str, 255, MF_BYPOSITION);
    if StrPas(str) = '&SAMPLE MENU (TSubclass32)' then
      RemoveMenu(GetMenu(Subc1.WindowHandle), 0, MF_BYPOSITION)  //remove menu if Win98
  end;

  DrawMenuBar(Subc1.WindowHandle);
  Start.enabled := True;
  Stop.enabled := False;
end;

procedure TForm1.StopClick(Sender: TObject);
begin
  if not Subc1.start then exit;
  Subc1.start := False;
  Cleanup;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  lb1.clear;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if Subc1.Start then StopClick(nil);  //always a good idea to do this
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Subc1.OnMsgReceived := Subc1MsgReceived;   //reason this is assigned here rather than design time is because all the
  randomize;                                 //$IFDEF's prevent Delphi from detecting a compatible method to assign
end;                                         // (only necessary if trying to make a D2/D3/D4 compatible program)


//OnMsgReceived event of the subclassing control

{$IFDEF VER100}    //if D3
procedure TForm1.Subc1MsgReceived(const Wnd: Integer; var Msg: Cardinal;
   var wParam, lParam: Integer; const ProcessId: Integer; rmFlag: TWhen;
   var return: Integer; var CallDefWndproc: Boolean);
{$ENDIF}                                                //because of all the different types between Delphi
{$IFDEF VER90}   //if D2                                //versions, this is necessary to work in all versions
procedure TForm1.Subc1MsgReceived(const Wnd: Integer; var Msg: Cardinal;
   var wParam, lParam: Longint; const ProcessId: Integer; rmFlag: TWhen;
   var return: Longint; var CallDefWndproc: Boolean);
{$ENDIF}
{$IFDEF VER120}   //if D4
procedure TForm1.Subc1MsgReceived(const Wnd: HWND; var Msg: Cardinal;
   var wParam, lParam: Integer; const ProcessId: Cardinal; rmFlag: TWhen;
   var return: Integer; var CallDefWndproc: Boolean);
{$ENDIF}
begin
  lb1.items.add(Subc1.msgtostr(msg));
  lb1.perform(LB_SETTOPINDEX, lb1.items.count-1, 0);
  if IsIconic(Wnd) then exit;  //if window minimized, don't want to do any drawing

  case Msg of
    WM_DESTROY: begin     //subclass automatically stopped when WM_DESTROY is seen
                  Start.Enabled := True;
                  Stop.Enabled := False;
                end;

    WM_COMMAND: if hiword(wParam) = 0 then  //if menu message...
                  if loword(wParam) = ME.command then     //can't do a case with constants
                    MakeWindowHole(msEllipse)
                  else
                  if loword(wParam) = MR.command then
                    MakeWindowHole(msRect)
                  else
                  if (loword(wParam) = MT.command) then
                    MakeWindowHole(msTriangle)
                  else
                  if (loword(wParam) = MN.command) then
                    MakeWindowHole(msNone)
                  else
                  if (loword(wParam) = RS.command) then begin
                    Subc1.ForceStop;  //stop the subclass, NEVER set the Start property in this event
                    Cleanup;
                  end;

    WM_NCHITTEST:   //if mouse action occuring in the non-client area of the window
    begin
      SetButtonRect;  //update button rect

         //if the mouse action is occuring over the button we're drawing
      if PtInRect(btnrect, Point(loWord(lparam)-wrect.left, hiword(lparam)-wrect.top)) then begin
        return := HTMINE;        //have the window procedure return our custom button constant
        CallDefWndProc := False; //when changing return value, MUST set this to False
      end;
    end;

    WM_NCPAINT,
    WM_NCACTIVATE,
    WM_ACTIVATEAPP,
    WM_ACTIVATE,           //when these messages fire, redraw the button on the title bar
    WM_GETTEXT,
    WM_SETTEXT,
    WM_ERASEBKGND,
    WM_NCLBUTTONUP:
    begin
      DrawWindowButton(False);   //redraw the button
      if (msg = WM_NCLBUTTONUP) and (wParam = HTMINE) then begin  //if the left mouse button is pressed AND
        HolePunched := not HolePunched;                           //it's pressed ON our button area
        with wrect do
        if HolePunched then
          MakeWindowHole(msWeird)
        else
          MakeWindowHole(msNone);
      end;
    end;

    WM_NCLBUTTONDOWN,
    WM_NCLBUTTONDBLCLK: if wParam = HTMINE then DrawWindowButton(True);   //if button goes down, redraw the button
  end;
end;

end.
